from __future__ import division

from collections import namedtuple

import six
import numpy as np
from numpy.linalg import norm

from neupy import init
from neupy.utils import as_tuple
from neupy.algorithms import Kohonen
from neupy.exceptions import WeightInitializationError
from neupy.algorithms.associative.base import BaseAssociative
from neupy.core.properties import (BaseProperty, TypedListProperty,
                                   ChoiceProperty, NumberProperty,
                                   ParameterProperty, IntProperty)
from .randomized_pca import randomized_pca
from .neighbours import (find_step_scaler_on_rect_grid,
                         find_neighbours_on_rect_grid,
                         find_neighbours_on_hexagon_grid,
                         find_step_scaler_on_hexagon_grid)


__all__ = ('SOFM',)


def neg_euclid_distance(input_data, weight):
    """
    Negative Euclidian distance between input
    data and weight.

    Parameters
    ----------
    input_data : array-like
        Input dataset.

    weight : array-like
        Neural network's weights.

    Returns
    -------
    array-like
    """
    euclid_dist = norm(input_data.T - weight, axis=0)
    return -np.expand_dims(euclid_dist, axis=0)


def cosine_similarity(input_data, weight):
    """
    Cosine similarity between input data and weight.

    Parameters
    ----------
    input_data : array-like
        Input dataset.

    weight : array-like
        Neural network's weights.

    Returns
    -------
    array-like
    """
    norm_prod = norm(input_data) * norm(weight, axis=0)
    summated_data = np.dot(input_data, weight)
    cosine_dist = summated_data / norm_prod
    return np.reshape(cosine_dist, (1, weight.shape[1]))


def decay_function(value, epoch, reduction_rate):
    """
    Applies to the input value monothonical decay.

    Parameters
    ----------
    value : int, float

    epoch : int
        Current training iteration (epoch).

    reduction_rate : int
        The larger the value the slower decay

    Returns
    -------
    float
    """
    return value / (1 + epoch / reduction_rate)


class SOFMWeightParameter(ChoiceProperty):
    expected_type = as_tuple(ParameterProperty.expected_type, six.string_types)

    def __set__(self, instance, value):
        if isinstance(value, six.string_types):
            return super(ChoiceProperty, self).__set__(instance, value)
        return BaseProperty.__set__(self, instance, value)

    def __get__(self, instance, owner):
        choice_key = super(ChoiceProperty, self).__get__(instance, owner)

        if isinstance(choice_key, six.string_types):
            return self.choices[choice_key]

        return choice_key


def sample_data(data, features_grid):
    """
    Samples from the data number of rows specified in
    the ``n_outputs`` argument. In case if ``n_outputs > n_samples``
    then sample will be with replacement.

    Parameters
    ----------
    data : matrix ``(n_samples, n_features)``
        Matrix where each row is a data sample.

    n_outputs : int
        Number of samples in the output.

    Returns
    -------
    matrix ``(n_features, n_outputs)``
    """
    n_outputs = np.prod(features_grid)
    n_samples, n_features = data.shape

    with_replacement = n_samples < n_outputs
    indeces = np.random.choice(n_samples, n_outputs,
                               replace=with_replacement)

    return data[indeces].T


def linear_initialization(data, features_grid):
    """
    Linear weight initialization base on the randomized PCA.

    Parameters
    ----------
    data : 2d array-like

    features_grid : tuple
        Tuple that defines shape of the feature grid.

    Returns
    -------
    2d array-like
        Initialized weights
    """
    cols = features_grid[1]
    n_nodes = np.prod(features_grid)

    n_pca_components = 2
    coord = np.zeros((n_nodes, n_pca_components))

    for i in range(n_nodes):
        coord[i, 0] = int(i / cols)
        coord[i, 1] = int(i % cols)

    maximum = np.max(coord, axis=0)
    coord = 2 * (coord / maximum - 0.5)

    data_mean = np.mean(data, axis=0)
    data_std = np.std(data, axis=0)
    data = (data - data_mean) / data_std

    eigenvectors, eigenvalues = randomized_pca(data, n_pca_components)

    norms = np.sqrt(np.einsum('ij,ij->i', eigenvectors, eigenvectors))
    eigenvectors = ((eigenvectors.T / norms) * eigenvalues).T
    weight = data_mean + coord.dot(eigenvectors) * data_std

    return weight.T


class SOFM(Kohonen):
    """
    Self-Organizing Feature Map (SOFM or SOM).

    Notes
    -----
    - Training data samples should have normalized features.

    Parameters
    ----------
    {BaseAssociative.n_inputs}

    n_outputs : int or None
        Number of outputs. Parameter is optional in case if
        ``feature_grid`` was specified.

        .. code-block:: python

            if n_outputs is None:
                n_outputs = np.prod(feature_grid)

    learning_radius : int
        Parameter defines radius within which we consider all
        neurons as neighbours to the winning neuron. The bigger
        the value the more neurons will be updated after each
        iteration.

        The ``0`` values means that we don't update
        neighbour neurons.

        Defaults to ``0``.

    std : int, float
        Parameters controls learning rate for each neighbour.
        The further neighbour  neuron from the winning neuron
        the smaller that learning rate for it. Learning rate
        scales based on the factors produced by the normal
        distribution with center in the place of a winning
        neuron and standard deviation specified as a parameter.
        The learning rate for the winning neuron is always equal
        to the value specified in the ``step`` parameter and for
        neighbour neurons it's always lower.

        The bigger the value for this parameter the bigger
        learning rate for the neighbour neurons.

        Defaults to ``1``.

    features_grid : list, tuple, None
        Feature grid defines shape of the output neurons.
        The new shape should be compatible with the number
        of outputs. It means that the following condition
        should be true:

        .. code-block:: python

            np.prod(features_grid) == n_outputs

        SOFM implementation supports n-dimensional grids.
        For instance, in order to specify grid as cube instead of
        the regular rectangular shape we can set up options as
        the following:

        .. code-block:: python

            SOFM(
                ...
                features_grid=(5, 5, 5),
                ...
            )

        Defaults to ``(n_outputs, 1)``.

    grid_type : {{``rect``, ``hexagon``}}
        Defines connection type in feature grid. Type defines
        which neurons we will consider as closest to the winning
        neuron during the training.

        - ``rect`` - Connections between neurons will be organized
          in hexagonal grid.

        - ``hexagon`` - Connections between neurons will be organized
          in hexagonal grid. It works only for 1d or 2d grids.

        Defaults to ``rect``.

    distance : {{``euclid``, ``dot_product``, ``cos``}}
        Defines function that will be used to compute
        closest weight to the input sample.

        - ``dot_product``: Just a regular dot product between
          data sample and network's weights

        - ``euclid``: Euclidian distance between data sample
          and network's weights

        - ``cos``: Cosine distance between data sample and
          network's weights

        Defaults to ``euclid``.

    reduce_radius_after : int or None
        Every specified number of epochs ``learning_radius``
        parameter will be reduced by ``1``. Process continues
        until ``learning_radius`` equal to ``0``.

        The ``None`` value disables parameter reduction
        during the training.

        Defaults to ``100``.

    reduce_step_after : int or None
        Defines reduction rate at which parameter ``step`` will
        be reduced using the following formula:

        .. code-block:: python

            step = step / (1 + current_epoch / reduce_step_after)

        The ``None`` value disables parameter reduction
        during the training.

        Defaults to ``100``.

    reduce_std_after : int or None
        Defines reduction rate at which parameter ``std`` will
        be reduced using the following formula:

        .. code-block:: python

            std = std / (1 + current_epoch / reduce_std_after)

        The ``None`` value disables parameter reduction
        during the training.

        Defaults to ``100``.

    weight : array-like, Initializer or {{``init_pca``, ``sample_from_data``}}
        Neural network weights.
        Value defined manualy should have shape ``(n_inputs, n_outputs)``.

        Also, it's possible to initialized weights base on the
        training data. There are two options:

        - ``sample_from_data`` - Before starting the training will
          randomly take number of training samples equal to number
          of expected outputs.

        - ``init_pca`` - Before training starts SOFM will applies PCA
          on a covariance matrix build from the training samples.
          Weights will be generated based on the two eigenvectors
          associated with the largest eigenvalues.

        Defaults to :class:`Normal() <neupy.init.Normal>`.

    {BaseNetwork.step}

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.epoch_end_signal}

    {BaseNetwork.train_end_signal}

    {Verbose.verbose}

    Methods
    -------
    init_weights(train_data)
        Initialized weights besed on the input data. It works only
        for the `init_pca` and `sample_from_data` options. For other
        cases it will throw an error.

    {BaseSkeleton.predict}

    {BaseAssociative.train}

    {BaseSkeleton.fit}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms, environment
    >>>
    >>> environment.reproducible()
    >>>
    >>> data = np.array([
    ...     [0.1961, 0.9806],
    ...     [-0.1961, 0.9806],
    ...     [-0.5812, -0.8137],
    ...     [-0.8137, -0.5812],
    ... ])
    >>>
    >>> sofm = algorithms.SOFM(
    ...     n_inputs=2,
    ...     n_outputs=2,
    ...     step=0.1,
    ...     learning_radius=0
    ... )
    >>> sofm.train(data, epochs=100)
    >>> sofm.predict(data)
    array([[0, 1],
           [0, 1],
           [1, 0],
           [1, 0]])
    """
    n_outputs = IntProperty(minval=1, allow_none=True, default=None)
    weight = SOFMWeightParameter(
        default=init.Normal(),
        choices={
            'init_pca': linear_initialization,
            'sample_from_data': sample_data,
        }
    )

    features_grid = TypedListProperty(allow_none=True, default=None)

    DistanceParameter = namedtuple('DistanceParameter', 'name func')
    distance = ChoiceProperty(
        default='euclid',
        choices={
            'dot_product': DistanceParameter(
                name='dot_product',
                func=np.dot),
            'euclid': DistanceParameter(
                name='euclid',
                func=neg_euclid_distance),
            'cos': DistanceParameter(
                name='cosine',
                func=cosine_similarity),
        })

    GridTypeMethods = namedtuple(
        'GridTypeMethods', 'name find_neighbours find_step_scaler')

    grid_type = ChoiceProperty(
        default='rect',
        choices={
            'rect': GridTypeMethods(
                name='rectangle',
                find_neighbours=find_neighbours_on_rect_grid,
                find_step_scaler=find_step_scaler_on_rect_grid),
            'hexagon': GridTypeMethods(
                name='hexagon',
                find_neighbours=find_neighbours_on_hexagon_grid,
                find_step_scaler=find_step_scaler_on_hexagon_grid)
        }
    )

    learning_radius = IntProperty(default=0, minval=0)
    std = NumberProperty(minval=0, default=1)

    reduce_radius_after = IntProperty(default=100, minval=1, allow_none=True)
    reduce_std_after = IntProperty(default=100, minval=1, allow_none=True)
    reduce_step_after = IntProperty(default=100, minval=1, allow_none=True)

    def __init__(self, **options):
        super(BaseAssociative, self).__init__(**options)

        if self.n_outputs is None and self.features_grid is None:
            raise ValueError("One of the following parameters has to be "
                             "specified: n_outputs, features_grid")

        elif self.n_outputs is None:
            self.n_outputs = np.prod(self.features_grid)

        n_grid_elements = np.prod(self.features_grid)
        invalid_feature_grid = (
            self.features_grid is not None and
            n_grid_elements != self.n_outputs)

        if invalid_feature_grid:
            raise ValueError(
                "Feature grid should contain the same number of elements "
                "as in the output layer: {0}, but found: {1} (shape: {2})"
                "".format(
                    self.n_outputs,
                    n_grid_elements,
                    self.features_grid))

        if self.features_grid is None:
            self.features_grid = (self.n_outputs, 1)

        if len(self.features_grid) > 2 and self.grid_type.name == 'hexagon':
            raise ValueError("SOFM with hexagon grid type should have "
                             "one or two dimensional feature grid, but got "
                             "{}d instead (shape: {!r})".format(
                                len(self.features_grid),
                                self.features_grid))

        is_pca_init = (
            isinstance(options.get('weight'), six.string_types) and
            options.get('weight') == 'init_pca')

        self.initialized = False
        if not callable(self.weight):
            self.init_layers()
            self.initialized = True

            if self.distance.name == 'cosine':
                self.weight /= np.linalg.norm(self.weight, axis=0)

        elif is_pca_init and self.grid_type.name != 'rectangle':
            raise WeightInitializationError(
                "Cannot apply PCA weight initialization for non-rectangular "
                "grid. Grid type: {}".format(self.grid_type.name))

    def predict_raw(self, input_data):
        input_data = self.format_input_data(input_data)

        n_samples = input_data.shape[0]
        output = np.zeros((n_samples, self.n_outputs))

        for i, input_row in enumerate(input_data):
            output[i, :] = self.distance.func(
                input_row.reshape(1, -1), self.weight)

        return output

    def update_indexes(self, layer_output):
        neuron_winner = layer_output.argmax(axis=1).item(0)
        winner_neuron_coords = np.unravel_index(
            neuron_winner, self.features_grid)

        learning_radius = self.learning_radius
        step = self.step
        std = self.std

        if self.reduce_radius_after is not None:
            learning_radius -= self.last_epoch // self.reduce_radius_after
            learning_radius = max(0, learning_radius)

        if self.reduce_step_after is not None:
            step = decay_function(step, self.last_epoch,
                                  self.reduce_step_after)

        if self.reduce_std_after is not None:
            std = decay_function(std, self.last_epoch,
                                 self.reduce_std_after)

        methods = self.grid_type
        output_grid = np.reshape(layer_output, self.features_grid)

        output_with_neightbours = methods.find_neighbours(
            grid=output_grid,
            center=winner_neuron_coords,
            radius=learning_radius)

        step_scaler = methods.find_step_scaler(
            grid=output_grid,
            center=winner_neuron_coords,
            std=std)

        index_y, = np.nonzero(
            output_with_neightbours.reshape(self.n_outputs))

        step_scaler = step_scaler.reshape(self.n_outputs)
        return index_y, step * step_scaler[index_y]

    def init_weights(self, input_train):
        if self.initialized:
            raise WeightInitializationError(
                "Weights have been already initialized")

        weight_initializer = self.weight
        self.weight = weight_initializer(input_train, self.features_grid)
        self.initialized = True

        if self.distance.name == 'cosine':
            self.weight /= np.linalg.norm(self.weight, axis=0)

    def train(self, input_train, summary='table', epochs=100):
        if not self.initialized:
            self.init_weights(input_train)

        super(SOFM, self).train(input_train, summary=summary, epochs=epochs)

    def train_epoch(self, input_train, target_train=None):
        step = self.step
        predict = self.predict
        update_indexes = self.update_indexes

        error = 0
        for input_row in input_train:
            input_row = np.reshape(input_row, (1, input_row.size))
            layer_output = predict(input_row)

            index_y, step = update_indexes(layer_output)
            distance = input_row.T - self.weight[:, index_y]
            updated_weights = (self.weight[:, index_y] + step * distance)

            if self.distance.name == 'cosine':
                updated_weights /= np.linalg.norm(updated_weights, axis=0)

            self.weight[:, index_y] = updated_weights
            error += np.abs(distance).mean()

        return error / len(input_train)



import os
import sys


__all__ = ('red', 'green', 'gray', 'white', 'bold', 'underline')


def is_color_supported():
    """
    Returns ``True`` if the running system's terminal supports
    color, and ``False`` otherwise.

    Notes
    -----
    Code from Djano: https://github.com/django/django/blob/\
    master/django/core/management/color.py

    Returns
    -------
    bool
    """
    supported_platform = (
        sys.platform != 'Pocket PC' and
        (sys.platform != 'win32' or 'ANSICON' in os.environ)
    )

    # isatty is not always implemented
    is_a_tty = hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()
    is_support = (supported_platform and is_a_tty)
    return is_support


def create_style(ansi_code, use_bright_mode=False):
    """
    Create style based on ANSI code number.

    Parameters
    ----------
    ansi_code : int
        ANSI style code.

    Returns
    -------
    function
        Function that takes string argument and add ANDI styles
        if its possible.
    """
    def style(text):
        if is_color_supported():
            mode = int(use_bright_mode)
            return "\033[{};{}m{}\033[0m".format(mode, ansi_code, text)
        return text
    return style


red = create_style(ansi_code=31)
green = create_style(ansi_code=32)
gray = create_style(ansi_code=37)
white = create_style(ansi_code=37)
bold = create_style(ansi_code=1, use_bright_mode=True)
underline = create_style(ansi_code=4, use_bright_mode=True)



from itertools import product

import matplotlib.pyplot as plt
from neupy import algorithms, environment, init

from utils import plot_2d_grid, make_circle, make_elipse, make_square


plt.style.use('ggplot')
environment.reproducible()


if __name__ == '__main__':
    GRID_WIDTH = 4
    GRID_HEIGHT = 4

    datasets = [
        make_square(),
        make_circle(),
        make_elipse(corr=0.7),
    ]
    configurations = [{
        'weight_init': init.Uniform(0, 1),
        'title': 'Random uniform initialization',
    }, {
        'weight_init': 'sample_from_data',
        'title': 'Sampled from the data',
    }, {
        'weight_init': 'init_pca',
        'title': 'Initialize with PCA',
    }]

    plt.figure(figsize=(15, 15))
    plt.title("Compare weight initialization methods for SOFM")

    red, blue = ('#E24A33', '#348ABD')
    n_columns = len(configurations)
    n_rows = len(datasets)
    index = 1

    for data, conf in product(datasets, configurations):
        sofm = algorithms.SOFM(
            n_inputs=2,
            features_grid=(GRID_HEIGHT, GRID_WIDTH),

            verbose=True,
            shuffle_data=True,
            weight=conf['weight_init'],

            learning_radius=8,
            reduce_radius_after=5,

            std=2,
            reduce_std_after=5,

            step=0.3,
            reduce_step_after=5,
        )

        if not sofm.initialized:
            sofm.init_weights(data)

        plt.subplot(n_rows, n_columns, index)

        plt.title(conf['title'])
        plt.scatter(*data.T, color=blue, alpha=0.05)
        plt.scatter(*sofm.weight, color=red)

        weights = sofm.weight.reshape((2, GRID_HEIGHT, GRID_WIDTH))
        plot_2d_grid(weights, color=red)

        index += 1

    plt.show()



import copy
import types
from functools import wraps
from itertools import product
from contextlib import contextmanager

import six
import tensorflow as tf

from neupy.layers.utils import preformat_layer_shape
from neupy.utils import as_tuple, tensorflow_session
from .utils import join, is_sequential
from .graph import LayerGraph
from .inline import InlineConnection


__all__ = ('LayerConnection', 'BaseConnection', 'ParallelConnection',
           'parallel')


def clean_layer_references(graph, layer_references):
    """
    Using list of layers and layer names convert it to
    complete list of layer objects. Function try to find
    layer by specified name in the graph and replate it in the
    layer list to the related object.

    Parameters
    ----------
    graph : LayerGraph instance
        Graph that has information about all layers from
        the ``layer_references`` list.

    layer_references : list, tuple
        List of layer instancens and layer names.

    Returns
    -------
    list of layers
    """
    layers = []

    for layer_reference in layer_references:
        if isinstance(layer_reference, six.string_types):
            layer_reference = graph.find_layer_by_name(layer_reference)

        layers.append(layer_reference)

    return layers


def check_initialization(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        result = method(*args, **kwargs)
        self.initialized = True
        return result
    return wrapper


def create_input_variables(input_layers):
    """
    Create input variables for each input layer
    in the graph.

    Parameters
    ----------
    input_layers : list of layers

    Returns
    -------
    list of Tensorflow variables
    """
    inputs = []

    for input_layer in input_layers:
        variable = tf.placeholder(
            tf.float32,
            shape=as_tuple(None, input_layer.output_shape),
            name="network-input/to-layer-{}".format(input_layer.name),
        )
        inputs.append(variable)

    return inputs


class BaseConnection(InlineConnection):
    """
    Base class from chain connections.

    Attributes
    ----------
    graph : LayerGraph
        Reference to the graph that contains all lations
        between layers specified in the connection.

    training_state : bool
        Defines state of the connection. Defaults to ``True``.

    input_layers : list of layers
        List of connection's input layers.

    output_layers : list of layers
        List of connection's output layers.
    """
    computation_cache = {}

    def __init__(self):
        self.initialized = False

        # Make sure that we save information when connection was
        # initialized. It will work even if method was reinitialized
        self.initialize = types.MethodType(
            check_initialization(self.initialize), self)

        self.training_state = True
        self.look_inside = 0
        self.graph = LayerGraph()

        self.input_layers = [self]
        self.output_layers = [self]

    def connect(self, left, right):
        """
        Make connection between two objects.
        """
        return LayerConnection(left, right)

    def __rshift__(self, other):
        return super(BaseConnection, self).__gt__(other)

    def __lshift__(self, other):
        return super(BaseConnection, self).__lt__(other)

    def __iter__(self):
        yield self

    def output(self, input_value):
        """
        Return output base on the input value.

        Parameters
        ----------
        input_value
        """
        raise NotImplementedError

    def initialize(self):
        """
        Initialize connection
        """

    @contextmanager
    def disable_training_state(self):
        """
        Disable connection's training state
        """
        self.training_state = False
        yield
        self.training_state = True

    def predict(self, *inputs):
        """
        Using current tensorflow session this method propagates
        input throught the network and returns output from it.
        """
        session = tensorflow_session()
        # We cache it in order to avoid graph creation
        # every time user calls prediction.
        cache_key = (session, id(self))

        if cache_key not in self.computation_cache:
            input_variables = create_input_variables(self.input_layers)

            with self.disable_training_state():
                self.computation_cache[cache_key] = {
                    'inputs': input_variables,
                    'outputs': self.output(*input_variables),
                }

        graph = self.computation_cache[cache_key]
        feed_dict = dict(zip(graph['inputs'], inputs))

        return session.run(graph['outputs'], feed_dict=feed_dict)


def make_common_graph(left_layer, right_layer):
    """
    Makes common graph for two layers that exists
    in different graphs.

    Parameters
    ----------
    left_layer : layer
    right_layer : layer

    Returns
    -------
    LayerGraph instance
        Graph that contains both layers and their connections.
    """
    graph = LayerGraph.merge(left_layer.graph, right_layer.graph)

    for layer in graph.forward_graph.keys():
        layer.graph = graph

    return graph


def topological_sort(graph):
    """
    Repeatedly go through all of the nodes in the graph, moving each of
    the nodes that has all its edges resolved, onto a sequence that
    forms our sorted graph. A node has all of its edges resolved and
    can be moved once all the nodes its edges point to, have been moved
    from the unsorted graph onto the sorted one.

    Parameters
    ----------
    graph : dict
        Dictionary that has graph structure.

    Raises
    ------
    RuntimeError
        If graph has cycles.

    Returns
    -------
    list
        List of nodes sorted in topological order.
    """
    sorted_nodes = []
    graph_unsorted = graph.copy()

    while graph_unsorted:
        acyclic = False

        for node, edges in list(graph_unsorted.items()):
            if all(edge not in graph_unsorted for edge in edges):
                acyclic = True
                del graph_unsorted[node]
                sorted_nodes.append(node)

        if not acyclic:
            raise RuntimeError("A cyclic dependency occurred")

    return sorted_nodes


class ParallelConnection(BaseConnection):
    """
    Connection between separate layer networks in parallel.
    Each network has it's own input and out layers.

    Parameters
    ----------
    connections : list of list or connection
        List of networks.

    Attributes
    ----------
    connections : list of connections
        List of networks.

    input_layers : list of connections
        Combined list of all input layers that each
        parallel connection has.

    output_layers : list of connections
        Combined list of all output layers that each
        parallel connection has.
    """
    def __init__(self, connections):
        from neupy.layers.base import ResidualConnection

        super(ParallelConnection, self).__init__()

        self.connections = []
        self.input_layers = []
        self.output_layers = []

        for layers in connections:
            if isinstance(layers, BaseConnection):
                connection = layers
            elif not layers:
                connection = ResidualConnection()
            else:
                connection = join(*layers)

            self.connections.append(connection)

            for input_layer in connection.input_layers:
                if input_layer not in self.input_layers:
                    self.input_layers.append(input_layer)

            for output_layer in connection.output_layers:
                if output_layer not in self.output_layers:
                    self.output_layers.append(output_layer)

    @property
    def input_shape(self):
        """
        Returns input shape per each network parallel
        connection.
        """
        input_shapes = []
        for connection in self.connections:
            input_shapes.append(connection.input_shape)
        return input_shapes

    @property
    def output_shape(self):
        """
        Returns output shape per each network parallel
        connection.
        """
        output_shapes = []
        for connection in self.connections:
            output_shapes.append(connection.output_shape)
        return output_shapes

    def initialize(self):
        """
        Initialize all connections.
        """
        for connection in self.connections:
            connection.initialize()

    def output(self, first_input, *other_inputs):
        """
        Compute outputs per each network in parallel
        connection.

        Parameters
        ----------
        first_input : Tensorfow variable, array-like, dict
        *other_inputs

        Returns
        -------
        list
        """
        n_inputs = len(other_inputs) + 1  # +1 for first input
        n_connections = len(self.connections)

        if not other_inputs:
            input_values = [first_input] * n_connections

        elif n_inputs == n_connections:
            input_values = as_tuple(first_input, other_inputs)

        else:
            raise ValueError("Expected {} input values for parallel "
                             "connection, got {}"
                             "".format(n_connections, n_inputs))

        outputs = []
        for input_value, connection in zip(input_values, self.connections):
            connection_output = connection.output(input_value)
            outputs.append(connection_output)

        return outputs

    @contextmanager
    def disable_training_state(self):
        """
        Disable training state for all layers in all
        connections.
        """
        for connection in self.connections:
            for layer in connection:
                layer.training_state = False

        yield

        for connection in self.connections:
            for layer in connection:
                layer.training_state = True

    def __iter__(self):
        for connection in self.connections:
            yield connection


def parallel(*connections):
    return ParallelConnection(connections)


class LayerConnection(BaseConnection):
    """
    Make connection between layers.

    Parameters
    ----------
    left : layer, connection or list of connections
    right : layer, connection or list of connections

    Attributes
    ----------
    input_layers : list
        List of input layers.

    input_shape : tuple or list of tuples
        Returns input shape as a tuple in case if network
        has only one input layer and list of tuples otherwise.

    output_layers : list
        List of output layers.

    output_shape : tuple or list of tuples
        Returns output shape as a tuple in case if network
        has only one output layer and list of tuples otherwise.

    layers : list
        Topologicaly sorted list of all layers.

    graph : LayerGraph instance
        Graph that stores relations between layer
        in the network.
    """
    def __init__(self, left, right):
        super(LayerConnection, self).__init__()

        self.left_raw = left
        self.right_raw = right

        if isinstance(left, (list, tuple)):
            left = ParallelConnection(left)

        if isinstance(right, (list, tuple)):
            right = ParallelConnection(right)

        self.left = left
        self.right = right

        layers = product(left.output_layers, right.input_layers)
        for left_output, right_input in layers:
            self.full_graph = make_common_graph(left_output, right_input)

        self.full_graph.connect_layers(left.output_layers, right.input_layers)

        self.input_layers = self.left.input_layers
        self.output_layers = self.right.output_layers

        # Generates subgraph that contains only connections
        # between specified input and output layers
        self.graph = self.full_graph.subgraph(
            self.input_layers, self.output_layers)

    @property
    def input_shape(self):
        """
        Connection's input shape/shapes.

        Returns
        -------
        list of tuples, tuple or None
            - List of tuples: in case if there are more than
              one input layer. Each tuple is a shape of the
              specific input layer.

            - tuple: in case if there is one input layer.
              Tuple object defines input layer's shape.
        """
        # Cannot save them during initialization step,
        # because input shape can be modified later
        if len(self.input_layers) == 1:
            input_layer = self.input_layers[0]
            return input_layer.input_shape

        input_shapes = []
        for input_layer in self.input_layers:
            input_shapes.append(input_layer.input_shape)

        return input_shapes

    @property
    def output_shape(self):
        """
        Connection's output shape/shapes.

        Returns
        -------
        list of tuples, tuple or None
            - List of tuples: in case if there are more than
              one input layer. Each tuple is a shape of the
              specific input layer.

            - tuple: in case if there is one input layer.
              Tuple object defines input layer's shape.
        """
        # Cannot save them during initialization step,
        # because input shape can be modified later
        if len(self.output_layers) == 1:
            output_layer = self.output_layers[0]
            return output_layer.output_shape

        output_shapes = []
        for output_layer in self.output_layers:
            output_shapes.append(output_layer.output_shape)

        return output_shapes

    def initialize(self):
        """
        Initialize all layers in the connection.
        """
        for layer in self:
            layer.initialize()

    def output(self, first_input, *other_inputs):
        """
        Propagate input values through all layers in the
        connections and returns output from the final layers.

        Parameters
        ----------
        first_input : Tensorfow variable, array-like, dict
            - Input values can be Tensorfow variables or
              array-like objects

            - Dictionary inputs should have key that
              define input layer and value is a variables
              that needs to be propagated through all layers.

        *other_inputs
            Suitable in case if we need to set up multiple
            input variables in a sequence.

        Returns
        -------
        Tensorfow expression
        """
        if other_inputs:
            input_values = as_tuple(first_input, other_inputs)
        else:
            input_values = first_input

        if isinstance(input_values, (list, tuple)):
            n_input_layers = len(self.input_layers)
            n_input_vars = len(input_values)

            if n_input_vars != n_input_layers:
                raise ValueError("Connection has {} input layer(s), "
                                 "but {} inputs was provided"
                                 "".format(n_input_layers, n_input_vars))

            # Layers in the self.graph.input_layers and
            # self.input_layers variables can have a different order.
            # Order in the self.input_layers is defined by user
            input_values_as_dict = {}

            for layer, value in zip(self.input_layers, input_values):
                input_values_as_dict[layer] = value

            input_values = input_values_as_dict

        return self.graph.propagate_forward(input_values)

    def start(self, first_layer, *other_layers):
        """
        Create new LayerConnection instance that point to
        different input layers.

        Parameters
        ----------
        first_layer : layer, str
            Layer instance or layer name.

        *other_layers
            Layer instances or layer names.

        Returns
        -------
        connection
        """
        input_layers = as_tuple(first_layer, other_layers)
        input_layers = clean_layer_references(self.graph, input_layers)

        subgraph = self.graph.subgraph_for_input(input_layers)

        new_connection = copy.copy(self)
        new_connection.graph = subgraph
        new_connection.input_layers = subgraph.input_layers

        # don't care about self.left and self.right attributes.
        # remove them to make sure that other function
        # won't use invalid references
        if hasattr(new_connection, 'left'):
            del new_connection.left

        if hasattr(new_connection, 'right'):
            del new_connection.right

        return new_connection

    def end(self, first_layer, *other_layers):
        """
        Create new LayerConnection instance that point to
        different output layers.

        Parameters
        ----------
        first_layer : layer, str
            Layer instance or layer name.

        *other_layers
            Layer instances or layer names.

        Returns
        -------
        connection
        """
        output_layers = as_tuple(first_layer, other_layers)
        output_layers = clean_layer_references(self.graph, output_layers)

        subgraph = self.graph.subgraph_for_output(output_layers)

        new_connection = copy.copy(self)
        new_connection.graph = subgraph
        new_connection.output_layers = subgraph.output_layers

        # don't care about self.left and self.right attributes.
        # remove them to make sure that other function
        # won't use invalid references
        if hasattr(new_connection, 'left'):
            del new_connection.left

        if hasattr(new_connection, 'right'):
            del new_connection.right

        return new_connection

    @property
    def layers(self):
        return list(self)

    def layer(self, layer_name):
        """
        Find layer instance in the network based on the
        specified layer name.

        Parameters
        ----------
        layer_name : str
            Name of the layer.

        Returns
        -------
        layer
        """
        return self.graph.find_layer_by_name(layer_name)

    @contextmanager
    def disable_training_state(self):
        """
        Disable training state for all layers in the
        connection.
        """
        for layer in self:
            layer.training_state = False

        yield

        for layer in self:
            layer.training_state = True

    def __len__(self):
        return len(self.graph.forward_graph)

    def __iter__(self):
        for layer in topological_sort(self.graph.backward_graph):
            yield layer

    def __repr__(self):
        n_layers = len(self)

        if n_layers > 5 or not is_sequential(self):
            return '{} -> [... {} layers ...] -> {}'.format(
                preformat_layer_shape(self.input_shape),
                n_layers,
                preformat_layer_shape(self.output_shape))

        return ' > '.join([repr(layer) for layer in self])



import tensorflow as tf

from neupy import init
from neupy.core.properties import (NumberProperty, ProperFractionProperty,
                                   ParameterProperty, IntProperty)
from neupy.utils import asfloat, as_tuple
from neupy.exceptions import LayerConnectionError
from .activations import AxesProperty
from .base import BaseLayer


__all__ = ('BatchNorm', 'LocalResponseNorm')


def find_opposite_axes(axes, ndim):
    """
    Based on the total number of dimensions function
    finds all axes that are missed in the specified
    list ``axes``.

    Parameters
    ----------
    axes : list or tuple
        Already known axes.

    ndim : int
        Total number of dimensions.

    Returns
    -------
    list

    Examples
    --------
    >>> from neupy.layers.normalization import find_opposite_axes
    >>> find_opposite_axes([0, 1], ndim=4)
    [2, 3]
    >>>
    >>> find_opposite_axes([], ndim=4)
    [0, 1, 2, 3]
    >>>
    >>> find_opposite_axes([0, 1, 2], ndim=3)
    []
    """
    if any(axis >= ndim for axis in axes):
        raise ValueError("Some axes have invalid values. Axis value "
                         "should be between 0 and {}".format(ndim))

    return [axis for axis in range(ndim) if axis not in axes]


class BatchNorm(BaseLayer):
    """
    Batch-normalization layer.

    Parameters
    ----------
    axes : int, tuple with int or None
        The axis or axes along which normalization is applied.
        ``None`` means that normalization will be applied over
        all axes except the first one. In case of 4D tensor it will
        be equal to ``(0, 1, 2)``. Defaults to ``None``.

    epsilon : float
        Epsilon is a positive constant that adds to the standard
        deviation to prevent the division by zero.
        Defaults to ``1e-5``.

    alpha : float
        Coefficient for the exponential moving average of
        batch-wise means and standard deviations computed during
        training; the closer to one, the more it will depend on
        the last batches seen. Value needs to be between ``0`` and ``1``.
        Defaults to ``0.1``.

    gamma : array-like, Tensorfow variable, scalar or Initializer
        Default initialization methods you can
        find :ref:`here <init-methods>`.
        Defaults to ``Constant(value=1)``.

    beta : array-like, Tensorfow variable, scalar or Initializer
        Default initialization methods you can
        find :ref:`here <init-methods>`.
        Defaults to ``Constant(value=0)``.

    running_mean : array-like, Tensorfow variable, scalar or Initializer
        Default initialization methods you can
        find :ref:`here <init-methods>`.
        Defaults to ``Constant(value=0)``.

    running_inv_std : array-like, Tensorfow variable, scalar or Initializer
        Default initialization methods you can
        find :ref:`here <init-methods>`.
        Defaults to ``Constant(value=1)``.

    {BaseLayer.Parameters}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    References
    ----------
    .. [1] Batch Normalization: Accelerating Deep Network Training
           by Reducing Internal Covariate Shift,
           http://arxiv.org/pdf/1502.03167v3.pdf
    """
    axes = AxesProperty(default=None)
    epsilon = NumberProperty(default=1e-5, minval=0)
    alpha = ProperFractionProperty(default=0.1)
    beta = ParameterProperty(default=init.Constant(value=0))
    gamma = ParameterProperty(default=init.Constant(value=1))

    running_mean = ParameterProperty(default=init.Constant(value=0))
    running_inv_std = ParameterProperty(default=init.Constant(value=1))

    def initialize(self):
        super(BatchNorm, self).initialize()

        input_shape = as_tuple(None, self.input_shape)
        ndim = len(input_shape)

        if self.axes is None:
            # If ndim == 4 then axes = (0, 1, 2)
            # If ndim == 2 then axes = (0,)
            self.axes = tuple(range(ndim - 1))

        if any(axis >= ndim for axis in self.axes):
            raise ValueError("Cannot apply batch normalization on the axis "
                             "that doesn't exist.")

        opposite_axes = find_opposite_axes(self.axes, ndim)
        parameter_shape = [
            input_shape[axis] if axis in opposite_axes else 1
            for axis in range(ndim)
        ]

        if any(parameter is None for parameter in parameter_shape):
            unknown_dim_index = parameter_shape.index(None)
            raise ValueError("Cannot apply batch normalization on the axis "
                             "with unknown size over the dimension #{} "
                             "(0-based indeces).".format(unknown_dim_index))

        self.add_parameter(value=self.running_mean, shape=parameter_shape,
                           name='running_mean', trainable=False)
        self.add_parameter(value=self.running_inv_std, shape=parameter_shape,
                           name='running_inv_std', trainable=False)

        self.add_parameter(value=self.gamma, name='gamma',
                           shape=parameter_shape, trainable=True)
        self.add_parameter(value=self.beta, name='beta',
                           shape=parameter_shape, trainable=True)

    def output(self, input_value):
        alpha = asfloat(self.alpha)
        running_mean = self.running_mean
        running_inv_std = self.running_inv_std

        if not self.training_state:
            mean, inv_std = running_mean, running_inv_std
        else:
            mean = tf.reduce_mean(
                input_value, self.axes,
                keepdims=True, name="mean",
            )
            variance = tf.reduce_mean(
                tf.squared_difference(input_value, tf.stop_gradient(mean)),
                self.axes,
                keepdims=True,
                name="variance",
            )
            inv_std = tf.rsqrt(variance + asfloat(self.epsilon))

            self.updates = [(
                running_inv_std,
                asfloat(1 - alpha) * running_inv_std + alpha * inv_std
            ), (
                running_mean,
                asfloat(1 - alpha) * running_mean + alpha * mean
            )]

        normalized_value = (input_value - mean) * inv_std
        return self.gamma * normalized_value + self.beta


class LocalResponseNorm(BaseLayer):
    """
    Local Response Normalization Layer.

    Aggregation is purely across channels, not within channels,
    and performed "pixelwise".

    If the value of the :math:`i` th channel is :math:`x_i`, the output is

    .. math::
        x_i = \\frac{{x_i}}{{ (k + ( \\alpha \\sum_j x_j^2 ))^\\beta }}

    where the summation is performed over this position on :math:`n`
    neighboring channels.

    Parameters
    ----------
    alpha : float
        Coefficient, see equation above

    beta : float
        Offset, see equation above

    k : float
        Exponent, see equation above

    depth_radius : int
        Number of adjacent channels to normalize over, must be odd.

    {BaseLayer.Parameters}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}
    """
    alpha = NumberProperty(default=1e-4)
    beta = NumberProperty(default=0.75)
    k = NumberProperty(default=2)
    depth_radius = IntProperty(default=5)

    def __init__(self, **options):
        super(LocalResponseNorm, self).__init__(**options)

        if self.depth_radius % 2 == 0:
            raise ValueError("Only works with odd ``n``")

    def validate(self, input_shape):
        ndim = len(input_shape)

        if ndim != 3:
            raise LayerConnectionError(
                "Layer `{}` expected input with 3 dimensions, got {}"
                "".format(self, ndim))

    def output(self, input_value):
        return tf.nn.local_response_normalization(
            input_value,
            depth_radius=self.depth_radius,
            bias=self.k,
            alpha=self.alpha,
            beta=self.beta,
        )



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle

from neupy.utils import format_data


__all__ = ('hinton',)


def hinton(matrix, max_weight=None, ax=None, add_legend=True):
    """
    Draw Hinton diagram for visualizing a weight matrix.

    Parameters
    ----------
    matrix: array-like
        Matrix that you want to visualise using Hinton diagram.

    max_weight : float
        Maximum value of the matrix. If it's equal to ``None``
        than value would be calculated using the maximum from
        the matrix. Defaults to ``None``.

    ax : object
        Matplotlib Axes instantce. If value equal to ``None``
        then function generate the new Axes instance. Defaults
        to ``None``.

    Returns
    -------
    object
        Matplotlib Axes instance.

    Examples
    --------
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> from neupy import plots
    >>>
    >>> weight = np.random.randn(20, 20)
    >>>
    >>> plt.style.use('ggplot')
    >>> plt.title("Hinton diagram")
    >>> plt.figure(figsize=(16, 12))
    >>> plots.hinton(weight)
    >>> plt.show()

    References
    ----------
    [1] http://matplotlib.org/examples/specialty_plots/hinton_demo.html
    """
    if ax is None:
        ax = plt.gca()

    matrix = format_data(matrix, is_feature1d=True)

    if max_weight is None:
        max_value = np.abs(matrix).max()
        max_value_log2_base = np.log(max_value) / np.log(2)
        max_weight = 2 ** np.ceil(max_value_log2_base)

    ax.patch.set_facecolor('gray')
    ax.set_aspect('equal', 'box')
    ax.xaxis.set_major_locator(plt.NullLocator())
    ax.yaxis.set_major_locator(plt.NullLocator())

    for (y, x), weight in np.ndenumerate(matrix):
        color = ('white' if weight > 0 else 'black')
        size = min(np.sqrt(np.abs(weight / max_weight)), 1.)
        rect = plt.Rectangle([x - size / 2., y - size / 2.], size, size,
                             facecolor=color, edgecolor=color)
        ax.add_patch(rect)

    ax.autoscale_view()
    ax.invert_yaxis()

    if add_legend:
        max_value = matrix.max().round(2)
        min_value = matrix.min().round(2)

        white = Rectangle(xy=(0, 0), width=1., height=1., linewidth=1.,
                          linestyle='solid', facecolor='#ffffff')
        black = Rectangle(xy=(0, 0), width=1., height=1., color='#000000')

        if min_value < 0 and max_value > 0:
            rectangles = [white, black]
            rect_description = [
                'Positive value\n'
                'Max: {}'.format(max_value),
                'Negative value\n'
                'Min: {}'.format(min_value),
            ]

        elif min_value >= 0:
            rectangles = [white]
            rect_description = [
                'Positive value\n'
                'Min: {}\n'
                'Max: {}'.format(min_value, max_value),
            ]

        else:
            rectangles = [black]
            rect_description = [
                'Negative value\n'
                'Min: {}\n'
                'Max: {}'.format(min_value, max_value),
            ]

        ax.legend(rectangles, rect_description, loc='center left',
                  bbox_to_anchor=(1., 0.5))

    return ax



import inspect
from functools import wraps

import numpy as np
from scipy.sparse import issparse
import tensorflow as tf


__all__ = ('format_data', 'asfloat', 'AttributeKeyDict', 'preformat_value',
           'as_tuple', 'number_type', 'all_equal', 'class_method_name_scope',
           'tensorflow_session', 'tensorflow_eval', 'tf_repeat',
           'initialize_uninitialized_variables', 'function_name_scope')


number_type = (int, float, np.floating, np.integer)


def function_name_scope(function):
    """
    Decorator that wraps any function with the name score that has the
    same name as a function.
    """
    @wraps(function)
    def wrapper(*args, **kwargs):
        with tf.name_scope(function.__name__):
            return function(*args, **kwargs)
    return wrapper


def class_method_name_scope(method):
    """
    Decorator that wraps any method with the name score that has the
    same name as a method.
    """
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        with tf.name_scope(self.__class__.__name__):
            return method(*args, **kwargs)

    wrapper.original_method = method
    return wrapper


def format_data(data, is_feature1d=True, copy=False, make_float=True):
    """
    Transform data in a standardized format.

    Notes
    -----
    It should be applied to the input data prior to use in
    learning algorithms.

    Parameters
    ----------
    data : array-like
        Data that should be formated. That could be, matrix, vector or
        Pandas DataFrame instance.

    is_feature1d : bool
        Should be equal to ``True`` if input data is a vector that
        contains N samples with 1 feature each. Defaults to ``True``.

    copy : bool
        Defaults to ``False``.

    make_float : bool
        If `True` then input will be converted to float.
        Defaults to ``False``.

    Returns
    -------
    ndarray
        The same input data but transformed to a standardized format
        for further use.
    """
    if data is None or issparse(data):
        return data

    if make_float:
        data = asfloat(data)

    if not isinstance(data, np.ndarray) or copy:
        data = np.array(data, copy=copy)

    # Valid number of features for one or two dimensions
    n_features = data.shape[-1]

    if data.ndim == 1:
        data_shape = (n_features, 1) if is_feature1d else (1, n_features)
        data = data.reshape(data_shape)

    return data


def asfloat(value):
    """
    Convert variable to 32 bit float number.

    Parameters
    ----------
    value : matrix, ndarray, Tensorfow variable or scalar
        Value that could be converted to float type.

    Returns
    -------
    matrix, ndarray, Tensorfow variable or scalar
        Output would be input value converted to 32 bit float.
    """
    float_type = 'float32'

    if isinstance(value, (np.matrix, np.ndarray)):
        if value.dtype != np.dtype(float_type):
            return value.astype(float_type)

        return value

    elif isinstance(value, (tf.Tensor, tf.SparseTensor)):
        return tf.cast(value, tf.float32)

    elif issparse(value):
        return value

    float_x_type = np.cast[float_type]
    return float_x_type(value)


class AttributeKeyDict(dict):
    """
    Modified built-in Python ``dict`` class. That modification
    helps get and set values like attributes.

    Examples
    --------
    >>> attrdict = AttributeKeyDict()
    >>> attrdict
    {}
    >>> attrdict.test_key = 'test_value'
    >>> attrdict
    {'test_key': 'test_value'}
    >>> attrdict.test_key
    'test_value'
    """

    def __getattr__(self, attrname):
        return self[attrname]

    def __setattr__(self, attrname, value):
        self[attrname] = value

    def __delattr__(self, attrname):
        del self[attrname]

    def __reduce__(self):
        return (self.__class__, (dict(self),))


def preformat_value(value):
    """
    Function pre-format input value depending on it's type.

    Parameters
    ----------
    value : object

    Returns
    -------
    object
    """
    if inspect.isfunction(value) or inspect.isclass(value):
        return value.__name__

    elif isinstance(value, (list, tuple, set)):
        return [preformat_value(v) for v in value]

    elif isinstance(value, (np.ndarray, np.matrix)):
        return value.shape

    elif hasattr(value, 'default'):
        return value.default

    return value


def as_tuple(*values):
    """
    Convert sequence of values in one big tuple.

    Parameters
    ----------
    *values
        Values that needs to be combined in one big tuple.

    Returns
    -------
    tuple
        All input values combined in one tuple

    Examples
    --------
    >>> as_tuple(None, (1, 2, 3), None)
    (None, 1, 2, 3, None)
    >>>
    >>> as_tuple((1, 2, 3), (4, 5, 6))
    (1, 2, 3, 4, 5, 6)
    """
    cleaned_values = []
    for value in values:
        if isinstance(value, (tuple, list)):
            cleaned_values.extend(value)
        else:
            cleaned_values.append(value)
    return tuple(cleaned_values)


def all_equal(array):
    """
    Checks if all elements in the array are equal.

    Parameters
    ----------
    array : list, tuple

    Raises
    ------
    ValueError
        If input array is empty

    Returns
    -------
    bool
        `True` in case if all elements are equal and
        `False` otherwise.
    """
    if not array:
        raise ValueError("Array is empty")

    first_item = array[0]

    if any(item != first_item for item in array):
        return False

    return True


def tensorflow_session():
    if hasattr(tensorflow_session, 'cache'):
        session = tensorflow_session.cache

        if not session._closed:
            return session

    config = tf.ConfigProto(
        allow_soft_placement=True,
        inter_op_parallelism_threads=0,
        intra_op_parallelism_threads=0,
    )
    session = tf.Session(config=config)

    tensorflow_session.cache = session
    return session


def tensorflow_eval(value):
    session = tensorflow_session()
    initialize_uninitialized_variables()
    return session.run(value)


@function_name_scope
def flatten(value):
    return tf.reshape(value, [-1])


@function_name_scope
def outer(a, b):
    a = tf.expand_dims(a, 1)
    b = tf.expand_dims(b, 0)
    return tf.matmul(a, b)


@function_name_scope
def dot(a, b):
    return tf.tensordot(a, b, 1)


def get_variable_size(variable):
    size = 1
    for dimension in variable.shape:
        size *= int(dimension)
    return size


def initialize_uninitialized_variables():
    session = tensorflow_session()
    global_vars = tf.global_variables()
    is_not_initialized = session.run([
        tf.is_variable_initialized(var) for var in global_vars])

    not_initialized_vars = [
        v for (v, f) in zip(global_vars, is_not_initialized) if not f]

    if len(not_initialized_vars):
        session.run(tf.variables_initializer(not_initialized_vars))


def tf_repeat(tensor, repeats):
    """
    Repeat elements of an tensor. The same as ``numpy.repeat``.

    Parameters
    ----------
    input : tensor
    repeats: list, tuple
        Number of repeat for each dimension, length must be the
        same as the number of dimensions in input.

    Returns
    -------
    tensor
        Has the same type as input. Has the shape
        of ``tensor.shape * repeats``.
    """
    with tf.variable_scope("repeat"):
        expanded_tensor = tf.expand_dims(tensor, -1)
        multiples = as_tuple(1, repeats)
        tiled_tensor = tf.tile(expanded_tensor, multiples)
        return tf.reshape(tiled_tensor, tf.shape(tensor) * repeats)



from neupy.utils import asfloat
from neupy.core.properties import IntProperty
from .base import SingleStepConfigurable


__all__ = ('StepDecay',)


class StepDecay(SingleStepConfigurable):
    """
    Algorithm minimizes learning step monotonically after
    each iteration.

    .. math::

        \\alpha_{{t + 1}} = \\frac{{\\alpha_{{0}}}}\
            {{1 + \\frac{{t}}{{m}}}}

    where :math:`\\alpha` is a step, :math:`t` is an epoch number
    and :math:`m` is a ``reduction_freq`` parameter.

    Notes
    -----
    Step will be reduced faster when you have smaller training batches.

    Parameters
    ----------
    reduction_freq : int
        Parameter controls step redution frequency. The larger the
        value the slower step parameter decreases.

        For instance, if ``reduction_freq=100``
        and ``step=0.12`` then after ``100`` epochs ``step`` is
        going to be equal to ``0.06`` (which is ``0.12 / 2``),
        after ``200`` epochs ``step`` is going to be equal to
        ``0.04`` (which is ``0.12 / 3``) and so on.

        Defaults to ``100`` epochs.

    Warns
    -----
    {SingleStepConfigurable.Warns}

    Examples
    --------
    >>> from neupy import algorithms
    >>>
    >>> bpnet = algorithms.GradientDescent(
    ...     (2, 4, 1),
    ...     step=0.1,
    ...     reduction_freq=100,
    ...     addons=[algorithms.StepDecay]
    ... )
    >>>

    See Also
    --------
    :network:`SearchThenConverge`
    """
    reduction_freq = IntProperty(minval=1, default=100)

    def init_train_updates(self):
        updates = super(StepDecay, self).init_train_updates()
        epoch = self.variables.epoch
        step = self.variables.step

        step_update_condition = self.step / (
            1 + epoch / asfloat(self.reduction_freq)
        )
        updates.extend([
            (step, step_update_condition),
        ])
        return updates



from itertools import chain
from functools import reduce


def join(*connections):
    """
    Connect two layers.

    Parameters
    ----------
    *connections : layers or connections

    Returns
    -------
    connection
        Layers connected in a sequence.

    Examples
    --------
    >>> from neupy import layers
    >>> conn = layers.join(
    ...     layers.Input(784),
    ...     layers.Relu(500),
    ...     layers.Relu(300),
    ...     layers.Softmax(10),
    ... )
    """
    from .base import LayerConnection, ParallelConnection

    if not connections:
        return

    if len(connections) == 1:
        connection = connections[0]

        if isinstance(connection, (list, tuple)):
            return ParallelConnection(connection)

        return connection

    merged_connections = reduce(LayerConnection, connections)
    return merged_connections


def is_sequential(connection):
    """
    Check whether graph connection is a sequence.

    Parameters
    ----------
    connection : connection

    Returns
    -------
    bool
    """
    forward_graph_layers = connection.graph.forward_graph.values()
    backward_graph_layers = connection.graph.backward_graph.values()

    for layers in chain(forward_graph_layers, backward_graph_layers):
        if len(layers) >= 2:
            # One of the layers has multiple input
            # or output connections
            return False

    return True



import re
import types
from collections import OrderedDict

import six
import tensorflow as tf

from neupy import init
from neupy.core.config import Configurable
from neupy.core.properties import ParameterProperty, IntProperty, Property
from neupy.utils import asfloat, as_tuple, class_method_name_scope
from neupy.layers.connections import BaseConnection


__all__ = ('BaseLayer', 'ParameterBasedLayer', 'ResidualConnection')


def generate_layer_name(layer):
    """
    Generates unique name for layer.

    Parameters
    ----------
    layer : BaseLayer

    Returns
    -------
    str
    """
    cls = layer.__class__

    layer_id = cls.global_identifiers_map[cls]
    cls.global_identifiers_map[cls] += 1

    classname = cls.__name__

    if classname.isupper():
        layer_name = classname.lower()
    else:
        layer_name = re.sub(r'(?<!^)(?=[A-Z][a-z_])', '-', classname)

    return "{}-{}".format(layer_name.lower(), layer_id)


def create_shared_parameter(value, name, shape):
    """
    Creates NN parameter as Tensorfow variable.

    Parameters
    ----------
    value : array-like, Tensorfow variable, scalar or Initializer
        Default value for the parameter.

    name : str
        Shared variable name.

    shape : tuple
        Parameter's shape.

    Returns
    -------
    Tensorfow variable.
    """
    if isinstance(value, tf.Variable):
        return value

    if isinstance(value, init.Initializer):
        value = value.sample(shape)

    return tf.Variable(asfloat(value), name=name, dtype=tf.float32)


def initialize_layer(layer_class, kwargs, was_initialized):
    """
    We have a separate method for initialization, becase default
    __reduce__ functionality requires variables to be specified
    in order, which neupy doesn't support.
    """
    layer = layer_class(**kwargs)

    if was_initialized:
        layer.initialize()

    return layer


class BaseLayer(BaseConnection, Configurable):
    """
    Base class for all layers.

    Parameters
    ----------
    name : str or None
        Layer's identifier. If name is equal to ``None`` than name
        will be generated automatically. Defaults to ``None``.

    Methods
    -------
    disable_training_state()
        Swith off trainig state.

    initialize()
        Set up important configurations related to the layer.

    Attributes
    ----------
    input_shape : tuple
        Layer's input shape.

    output_shape : tuple
        Layer's output shape.

    training_state : bool
        Defines whether layer in training state or not.

    parameters : dict
        Trainable parameters.

    graph : LayerGraph instance
        Graphs that stores all relations between layers.
    """
    name = Property(expected_type=six.string_types)

    # Stores global identifier index for each layer class
    global_identifiers_map = {}

    def __new__(cls, *args, **kwargs):
        if cls not in cls.global_identifiers_map:
            cls.global_identifiers_map[cls] = 1
        return super(BaseLayer, cls).__new__(cls)

    def __init__(self, *args, **options):
        super(BaseLayer, self).__init__(*args)

        self.output = types.MethodType(
            class_method_name_scope(self.output), self)

        self.updates = []
        self.parameters = OrderedDict()
        self.name = generate_layer_name(layer=self)
        self.input_shape_ = None

        self.graph.add_layer(self)
        Configurable.__init__(self, **options)

    def validate(self, input_shape):
        """
        Validate input shape value before assigning it.

        Parameters
        ----------
        input_shape : tuple with int
        """

    @property
    def input_shape(self):
        return self.input_shape_

    @input_shape.setter
    def input_shape(self, shape):
        self.validate(shape)
        self.input_shape_ = shape

    @property
    def output_shape(self):
        return self.input_shape

    def output(self, input_value):
        return input_value

    def add_parameter(self, value, name, shape=None, trainable=True):
        layer_name = 'layer/{layer_name}/{parameter_name}'.format(
            layer_name=self.name,
            parameter_name=name.replace('_', '-'))

        parameter = create_shared_parameter(value, layer_name, shape)
        parameter.is_trainable = trainable

        self.parameters[name] = parameter

        setattr(self, name, parameter)
        return parameter

    def __repr__(self):
        classname = self.__class__.__name__
        return '{name}()'.format(name=classname)

    def __reduce__(self):
        parameters = self.get_params()
        return (initialize_layer, (
            self.__class__, parameters, self.initialized))


class ResidualConnection(BaseLayer):
    """
    Residual skip connection.
    """


class ParameterBasedLayer(BaseLayer):
    """
    Layer that creates weight and bias parameters.

    Parameters
    ----------
    size : int
        Layer's output size.

    weight : array-like, Tensorfow variable, scalar or Initializer
        Defines layer's weights. Default initialization methods
        you can find :ref:`here <init-methods>`.
        Defaults to :class:`XavierNormal() <neupy.init.XavierNormal>`.

    bias : 1D array-like, Tensorfow variable, scalar, Initializer or None
        Defines layer's bias.
        Default initialization methods you can find
        :ref:`here <init-methods>`. Defaults to
        :class:`Constant(0) <neupy.init.Constant>`.
        The ``None`` value excludes bias from the calculations and
        do not add it into parameters list.

    {BaseLayer.Parameters}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}
    """
    size = IntProperty(minval=1)
    weight = ParameterProperty(default=init.XavierNormal())
    bias = ParameterProperty(default=init.Constant(value=0), allow_none=True)

    def __init__(self, size, **options):
        super(ParameterBasedLayer, self).__init__(size=size, **options)

    @property
    def weight_shape(self):
        return as_tuple(self.input_shape, self.output_shape)

    @property
    def bias_shape(self):
        if self.bias is not None:
            return as_tuple(self.output_shape)

    def initialize(self):
        super(ParameterBasedLayer, self).initialize()

        self.add_parameter(value=self.weight, name='weight',
                           shape=self.weight_shape, trainable=True)

        if self.bias is not None:
            self.add_parameter(value=self.bias, name='bias',
                               shape=self.bias_shape, trainable=True)

    def __repr__(self):
        classname = self.__class__.__name__
        return '{name}({size})'.format(name=classname, size=self.size)



from neupy import layers
from neupy.utils import function_name_scope


__all__ = ('resnet50',)


@function_name_scope
def ResidualUnit(n_input_filters, n_output_filters, stride,
                 has_branch=False, name=None):

    def bn_name(index):
        return 'bn' + name + '_branch' + index

    def conv_name(index):
        return 'res' + name + '_branch' + index

    main_branch = layers.join(
        # The main purpose of this 1x1 covolutional layer is to
        # reduce number of filters. For instance, for the tensor with
        # 256 filters it can be reduced to 64 reducing amount of
        # computation by factor of 4.
        layers.Convolution(
            size=(1, 1, n_input_filters),
            stride=stride,
            bias=None,
            name=conv_name('2a'),
        ),
        layers.BatchNorm(name=bn_name('2a')),
        layers.Relu(),

        # This convolutional layer applies 3x3 filter in order to
        # extract features.
        layers.Convolution(
            (3, 3, n_input_filters),
            padding='SAME',
            bias=None,
            name=conv_name('2b'),
        ),
        layers.BatchNorm(name=bn_name('2b')),
        layers.Relu(),

        # last layer reverse operations of the first layer. In this
        # case we increase number of filters. For instamce, from previously
        # obtained 64 filters we can increase it back to the 256 filters
        layers.Convolution(
            (1, 1, n_output_filters),
            bias=None,
            name=conv_name('2c')
        ),
        layers.BatchNorm(name=bn_name('2c')),
    )

    if has_branch:
        residual_branch = layers.join(
            layers.Convolution(
                (1, 1, n_output_filters),
                stride=stride,
                bias=None,
                name=conv_name('1'),
            ),
            layers.BatchNorm(name=bn_name('1')),
        )
    else:
        # Empty list defines residual connection, meaning that
        # output from this branch would be equal to its input
        residual_branch = []

    return layers.join(
        # For the output from two branches we just combine results
        # with simple elementwise sum operation. The main purpose of
        # the residual connection is to build shortcuts for the
        # gradient during backpropagation.
        [main_branch, residual_branch],
        layers.Elementwise(),
        layers.Relu(),
    )


def resnet50():
    """
    ResNet50 network architecture with random parameters. Parameters
    can be loaded using ``neupy.storage`` module.

    ResNet50 has roughly 25.5 million parameters.

    Notes
    -----
    Because of the global pooling layer, ResNet50 can be applied to
    the images with variable sizes. The only limitation is that image
    size should be bigger than 32x32, otherwise network wont be able
    to apply all transformations to the image.

    Examples
    --------
    >>> from neupy import architectures
    >>> resnet50 = architectures.resnet50()
    >>> resnet50
    (224, 224, 3) -> [... 187 layers ...] -> 1000
    >>>
    >>> from neupy import algorithms
    >>> network = algorithms.Momentum(resnet50)

    See Also
    --------
    :architecture:`vgg16` : VGG16 network
    :architecture:`squeezenet` : SqueezeNet network
    :architecture:`resnet50` : ResNet-50 network

    References
    ----------
    Deep Residual Learning for Image Recognition.
    https://arxiv.org/abs/1512.03385
    """
    return layers.join(
        layers.Input((224, 224, 3)),

        # Convolutional layer reduces image's height and width by a factor
        # of 2 (because of the stride)
        # from (3, 224, 224) to (64, 112, 112)
        layers.Convolution((7, 7, 64), stride=2, bias=None,
                           padding='SAME', name='conv1'),

        layers.BatchNorm(name='bn_conv1'),
        layers.Relu(),

        # Stride equal two 2 reduces image size by a factor of two
        # from (64, 112, 112) to (64, 56, 56)
        layers.MaxPooling((3, 3), stride=2, padding="SAME"),

        # The branch option applies extrax convolution x+ batch
        # normalization transforamtions to the residual
        ResidualUnit(64, 256, stride=1, name='2a', has_branch=True),
        ResidualUnit(64, 256, stride=1, name='2b'),
        ResidualUnit(64, 256, stride=1, name='2c'),

        # Another stride=2 reduces width and hight by factor of 2
        ResidualUnit(128, 512, stride=2, name='3a', has_branch=True),
        ResidualUnit(128, 512, stride=1, name='3b'),
        ResidualUnit(128, 512, stride=1, name='3c'),
        ResidualUnit(128, 512, stride=1, name='3d'),

        # Another stride=2 reduces width and hight by factor of 2
        ResidualUnit(256, 1024, stride=2, name='4a', has_branch=True),
        ResidualUnit(256, 1024, stride=1, name='4b'),
        ResidualUnit(256, 1024, stride=1, name='4c'),
        ResidualUnit(256, 1024, stride=1, name='4d'),
        ResidualUnit(256, 1024, stride=1, name='4e'),
        ResidualUnit(256, 1024, stride=1, name='4f'),

        # Another stride=2 reduces width and hight by factor of 2
        ResidualUnit(512, 2048, stride=2, name='5a', has_branch=True),
        ResidualUnit(512, 2048, stride=1, name='5b'),
        ResidualUnit(512, 2048, stride=1, name='5c'),

        # Since the final residual unit has 2048 output filters, global
        # pooling will replace every output image with single average value.
        # Despite input iamge size output from this layer always will be
        # vector with 2048 values
        layers.GlobalPooling(),
        layers.Softmax(1000, name='fc1000'),
    )



from itertools import product

import numpy as np
import matplotlib.pyplot as plt


def iter_neighbours(weights, hexagon=False):
    _, grid_height, grid_width = weights.shape

    hexagon_even_actions = ((-1, 0), (0, -1), (1, 0), (0, 1), (1, 1), (-1, 1))
    hexagon_odd_actions = ((-1, 0), (0, -1), (1, 0), (0, 1), (-1, -1), (1, -1))
    rectangle_actions = ((-1, 0), (0, -1), (1, 0), (0, 1))

    for neuron_x, neuron_y in product(range(grid_height), range(grid_width)):
        neighbours = []

        if hexagon and neuron_x % 2 == 1:
            actions = hexagon_even_actions
        elif hexagon:
            actions = hexagon_odd_actions
        else:
            actions = rectangle_actions

        for shift_x, shift_y in actions:
            neigbour_x = neuron_x + shift_x
            neigbour_y = neuron_y + shift_y

            if 0 <= neigbour_x < grid_height and 0 <= neigbour_y < grid_width:
                neighbours.append((neigbour_x, neigbour_y))

        yield (neuron_x, neuron_y), neighbours


def plot_2d_grid(weights, ax=None, color='green', hexagon=False):
    if weights.ndim != 3:
        raise ValueError("Number of dimensions should be equal to 3 "
                         "(shape: (2, height, width)), got {} instead"
                         "".format(weights.ndim))

    n_features = weights.shape[0]

    if n_features != 2:
        raise ValueError("First dimension should be equal to 2")

    if ax is None:
        ax = plt.gca()

    for (neuron_x, neuron_y), neighbours in iter_neighbours(weights, hexagon):
        for (neigbour_x, neigbour_y) in neighbours:
            neurons_x_coords = (neuron_x, neigbour_x)
            neurons_y_coords = (neuron_y, neigbour_y)

            neurons = weights[:, neurons_x_coords, neurons_y_coords]
            ax.plot(*neurons, color=color)


def make_square():
    return np.random.random((10000, 2))


def make_circle():
    data = make_square()
    x, y = data[:, 0], data[:, 1]

    distance_from_center = ((x - 0.5) ** 2 + (y - 0.5) ** 2)
    return data[distance_from_center <= 0.5 ** 2]


def make_elipse(corr=0.8):
    projection = np.array([
        [corr, 1 - corr],
        [1 - corr, corr]])

    data = make_circle()
    return data.dot(projection)



import abc
import math

import six
import numpy as np

from neupy.core.docs import SharedDocsABCMeta


__all__ = ('Initializer', 'Constant', 'Normal', 'Uniform', 'Orthogonal',
           'HeNormal', 'HeUniform', 'XavierNormal', 'XavierUniform')


def identify_fans(shape):
    """
    Identify fans from shape.

    Parameters
    ----------
    shape : tuple or list
        Matrix shape.

    Returns
    -------
    tuple
        Tuple that contains :math:`fan_{in}` and :math:`fan_{out}`.
    """
    fan_in = shape[0]
    output_feature_shape = shape[1:]

    if output_feature_shape:
        fan_out = np.prod(output_feature_shape).item(0)
    else:
        fan_out = 1

    return fan_in, fan_out


def classname(instance):
    """
    Returns instance class name.

    Parameters
    ----------
    instance : object

    Returns
    -------
    str
    """
    return instance.__class__.__name__


class Initializer(six.with_metaclass(SharedDocsABCMeta)):
    """
    Base class for parameter initialization.

    Methods
    -------
    sample(shape)
        Returns tensor with specified shape.
    """
    inherit_method_docs = True

    @abc.abstractmethod
    def sample(self, shape):
        """
        Returns tensor with specified shape.

        Parameters
        ----------
        shape : tuple
            Parameter shape.

        Returns
        -------
        array-like
        """
        raise NotImplementedError

    def __repr__(self):
        return '{}()'.format(classname(self))


class Constant(Initializer):
    """
    Initialize parameter that has constant values.

    Parameters
    ----------
    value : float, int
        All parameters in the tensor will be equal to
        this value. Defaults to ``0``.

    Methods
    -------
    {Initializer.Methods}
    """
    def __init__(self, value=0):
        self.value = value

    def sample(self, shape):
        return np.ones(shape) * self.value

    def __repr__(self):
        return '{}({})'.format(classname(self), self.value)


class Normal(Initializer):
    """
    Initialize parameter sampling from the normal
    distribution.

    Parameters
    ----------
    mean : int, float
        Mean of the normal distribution.

    std : int, float
        Standard deviation of the normal distribution.

    Methods
    -------
    {Initializer.Methods}
    """
    def __init__(self, mean=0, std=0.01):
        self.mean = mean
        self.std = std

    def sample(self, shape):
        return np.random.normal(loc=self.mean, scale=self.std, size=shape)

    def __repr__(self):
        return '{}(mean={}, std={})'.format(
            classname(self), self.mean, self.std)


class Uniform(Initializer):
    """
    Initialize parameter sampling from the uniformal
    distribution.

    Parameters
    ----------
    minval : int, float
        Minimum possible value.

    maxval : int, float
        Maximum possible value.

    Methods
    -------
    {Initializer.Methods}
    """
    def __init__(self, minval=0, maxval=1):
        self.minval = minval
        self.maxval = maxval

    def sample(self, shape):
        minval, maxval = self.minval, self.maxval
        return np.random.random(shape) * (maxval - minval) + minval

    def __repr__(self):
        return '{}({}, {})'.format(classname(self),
                                   self.minval, self.maxval)


class Orthogonal(Initializer):
    """
    Initialize matrix with orthogonal basis.

    Parameters
    ----------
    scale : int, float
        Scales output matrix by a specified factor.
        Defaults to ``1``.

    Raises
    ------
    ValueError
        In case if tensor shape has more than 2
        dimensions.

    Methods
    -------
    {Initializer.Methods}
    """
    def __init__(self, scale=1):
        self.scale = scale

    def sample(self, shape):
        ndim = len(shape)

        if ndim not in (1, 2):
            raise ValueError("Shape attribute must have 1 or 2 dimensions. "
                             "Found {} dimensions".format(ndim))

        rand_matrix = np.random.randn(*shape)

        if ndim == 1:
            return rand_matrix

        nrows, ncols = shape
        u, _, v = np.linalg.svd(rand_matrix, full_matrices=False)
        ortho_base = u if nrows > ncols else v

        return self.scale * ortho_base[:nrows, :ncols]

    def __repr__(self):
        return '{}(scale={})'.format(classname(self), self.scale)


class InitializerWithGain(Initializer):
    """
    Initialization class that has gain property

    Parameters
    ----------
    gain : float or {{``relu``}}
        Multiplies scaling factor by speified gain.
        The ``relu`` values set up gain equal to :math:`\\sqrt{{2}}`.
        Defaults to ``1``.
    """
    def __init__(self, gain=1.0):
        if gain == 'relu':
            gain = math.sqrt(2)

        self.gain = gain
        super(InitializerWithGain, self).__init__()


class HeNormal(InitializerWithGain):
    """
    Kaiming He parameter initialization method based on the
    normal distribution.

    Parameters
    ----------
    {InitializerWithGain.Parameters}

    Methods
    -------
    {Initializer.Methods}

    References
    ----------
    [1] Kaiming He, Xiangyu Zhan, Shaoqing Ren, Jian Sun.
        Delving Deep into Rectifiers: Surpassing Human-Level
        Performance on ImageNet Classification, 2015.
    """
    def sample(self, shape):
        fan_in, _ = identify_fans(shape)
        variance = 2. / fan_in
        std = self.gain * np.sqrt(variance)
        return np.random.normal(loc=0, scale=std, size=shape)


class HeUniform(InitializerWithGain):
    """
    Kaiming He parameter initialization method based on the
    uniformal distribution.

    Parameters
    ----------
    {InitializerWithGain.Parameters}

    Methods
    -------
    {Initializer.Methods}

    References
    ----------
    [1] Kaiming He, Xiangyu Zhan, Shaoqing Ren, Jian Sun.
        Delving Deep into Rectifiers: Surpassing Human-Level
        Performance on ImageNet Classification, 2015.
    """
    def sample(self, shape):
        fan_in, _ = identify_fans(shape)
        variance = 6. / fan_in
        abs_max_value = self.gain * np.sqrt(variance)

        uniform = Uniform(minval=-abs_max_value, maxval=abs_max_value)
        return uniform.sample(shape)


class XavierNormal(InitializerWithGain):
    """
    Xavier Glorot parameter initialization method based on
    normal distribution.

    Parameters
    ----------
    {InitializerWithGain.Parameters}

    Methods
    -------
    {Initializer.Methods}

    References
    ----------
    [1] Xavier Glorot, Y Bengio. Understanding the difficulty
        of training deep feedforward neural networks, 2010.
    """
    def sample(self, shape):
        fan_in, fan_out = identify_fans(shape)
        variance = 2. / (fan_in + fan_out)
        std = self.gain * np.sqrt(variance)
        return np.random.normal(loc=0, scale=std, size=shape)


class XavierUniform(InitializerWithGain):
    """
    Xavier Glorot parameter initialization method based
    on uniform distribution.

    Methods
    -------
    {Initializer.Methods}

    References
    ----------
    [1] Xavier Glorot, Y Bengio. Understanding the difficulty
        of training deep feedforward neural networks, 2010.
    """
    def sample(self, shape):
        fan_in, fan_out = identify_fans(shape)
        variance = 6. / (fan_in + fan_out)
        abs_max_value = self.gain * np.sqrt(variance)

        uniform = Uniform(minval=-abs_max_value, maxval=abs_max_value)
        return uniform.sample(shape)



import numpy as np
import tensorflow as tf


__all__ = ('preformat_layer_shape', 'dimshuffle', 'iter_parameters',
           'count_parameters', 'extract_connection')


def preformat_layer_shape(shape):
    """
    Format layer's input or output shape.

    Parameters
    ----------
    shape : int or tuple

    Returns
    -------
    int or tuple
    """
    if isinstance(shape, tuple) and len(shape) == 1:
        return shape[0]
    return shape


def dimshuffle(value, ndim, axes):
    """
    Shuffle dimension based on the specified number of
    dimensions and axes.

    Parameters
    ----------
    value : Tensorfow variable
    ndim : int
    axes : tuple, list

    Returns
    -------
    Tensorfow variable
    """
    for dim in range(ndim):
        if dim not in axes:
            value = tf.expand_dims(value, dim)

    return value


def iter_parameters(layers, only_trainable=True):
    """
    Iterate through layer parameters.

    Parameters
    ----------
    layers : list of layers or connection

    only_trainable : bool
        If `True` returns only trainable parameters.
        Defaults to `True`.

    Yields
    ------
    tuple
        Tuple with three ariables: (layer, attribute_name, parameter)
    """
    observed_parameters = []

    for layer in layers:
        for attrname, param in layer.parameters.items():
            new_param = param not in observed_parameters

            if new_param and (param.is_trainable or not only_trainable):
                observed_parameters.append(param)
                yield layer, attrname, param


def count_parameters(connection):
    """
    Count number of parameters in Neural Network.

    Parameters
    ----------
    connection : list of laters or connection

    Returns
    -------
    int
        Number of parameters.
    """
    n_parameters = 0

    for _, _, parameter in iter_parameters(connection):
        shape = parameter.get_shape()
        n_parameters += np.prod(shape.as_list())

    return n_parameters


def extract_connection(instance):
    """
    Extract connection from different types of object.

    Parameters
    ----------
    instance : network, connection, list or tuple

    Returns
    -------
    connection

    Raises
    ------
    ValueError
        In case if input object doesn't have connection of layers.
    """
    # Note: Import it here in order to prevent loops
    from neupy.algorithms.base import BaseNetwork
    from neupy.layers.base import BaseConnection
    from neupy import layers

    if isinstance(instance, (list, tuple)):
        return layers.join(*instance)

    if isinstance(instance, BaseNetwork):
        return instance.connection

    if isinstance(instance, BaseConnection):
        return instance

    raise TypeError("Invalid input type. Input should be network, connection "
                    "or list of layers, got {}".format(type(instance)))
